"""
Utilities to generate chunk masks used in the Precomputed Connectome

"""

import os
import numpy as np
from tqdm import trange
from nilearn import image
from scipy.spatial.distance import cityblock


def get_distances(home_coords, locations):
    """Get Manhattan distances of voxels to home coordinates.

    Parameters
    ----------
    home_coords : (int, int, int)
        X/Y/Z coordinates of center of search region.
    locations : ndarray
        2D array of X/Y/Z coordinates of voxels.

    Returns
    -------
    distances : list of double
        Cityblock distances of voxels to home coordinates.

    """
    return [cityblock(home_coords, voxel) for voxel in locations]


def get_locations(volume):
    """Get coords of unassigned voxels in search area.

    Parameters
    ----------
    volume : ndarray
        Volume to search for unassigned voxels. Unassigned voxels have value 1.

    Returns
    -------
    locations : ndarray
        2D array of X/Y/Z coordinates of unassigned voxels.

    """
    return np.transpose(np.stack(np.where(volume)))


def get_extrema(radius_len, home_coords, XMAX, YMAX, ZMAX):
    """Get new boundaries of search area.

    Parameters
    ----------
    radius_len : int
        Radius to expand search area around home coordinates.
    home_coords : tuple of int
        X/Y/Z coordinates of center of search region.
    XMAX : int
        Maximum X-coordinate.
    YMAX : int
        Maximum Y-coordinate.
    ZMAX : int
        Maximum Z-coordinate.

    Returns
    -------
    xMin, xMax, yMin, yMax, zMin, zMax : tuple of int
        X/Y/Z coordinate boundaries of search region.

    """
    xMin = np.max([home_coords[0] - radius_len, 0])
    xMax = np.min([home_coords[0] + radius_len, XMAX - 1])
    yMin = np.max([home_coords[1] - radius_len, 0])
    yMax = np.min([home_coords[1] + radius_len, YMAX - 1])
    zMin = np.max([home_coords[2] - radius_len, 0])
    zMax = np.min([home_coords[2] + radius_len, ZMAX - 1])
    return xMin, xMax, yMin, yMax, zMin, zMax


def label_chunk(
    radius_len, home_coords, voxel_maps, chunk_label, chunk_size, XMAX, YMAX, ZMAX
):
    """Label voxels with chunk index.

    Parameters
    ----------
    radius_len : int
        Radius around home coordinates to search for unassigned voxels.
    home_coords : tuple of int
        X/Y/Z coordinates of home position.
    voxel_maps : ndarray
        Current state of brain volume voxel labels.
    chunk_label : int
        Chunk index to label assigned voxels with.
    chunk_size : int
        Number of voxels in a chunk.
    XMAX : int
        Maximum X-coordinate.
    YMAX : int
        Maximum Y-coordinate.
    ZMAX : int
        Maximum Z-coordinate.

    Returns
    -------
    voxel_maps : ndarray
        State of brain volume voxel labels after adding the latest chunk.

    """
    # Get boundaries of search area
    xMin, xMax, yMin, yMax, zMin, zMax = get_extrema(
        radius_len, home_coords, XMAX, YMAX, ZMAX
    )
    # Make a copy of the voxel_map
    search_area = voxel_maps.copy()
    # zero out everything outside the search area and zero out the home location
    search_area[:xMin, :, :] = 0
    search_area[xMax:, :, :] = 0
    search_area[:, :yMin, :] = 0
    search_area[:, yMax:, :] = 0
    search_area[:, :, :zMin] = 0
    search_area[:, :, zMax:] = 0
    search_area[home_coords[0], home_coords[1], home_coords[2]] = 0
    # Get voxelmap of unassigned voxels inside search area
    search_area = search_area == -1
    # Check if there are enough unassigned voxels to fill a chunk
    if np.sum(search_area) < (chunk_size - 1):
        label_chunk(
            radius_len + 1,
            home_coords,
            voxel_maps,
            chunk_label,
            chunk_size,
            XMAX,
            YMAX,
            ZMAX,
        )
    else:
        # Get coords of unassigned voxels in search area
        locations = get_locations(search_area)
        # Get Manhattan distances of unassigned voxels to home location
        distances = get_distances(home_coords, locations)
        for i in range(chunk_size - 1):
            minidx = np.argmin(distances)
            x, y, z = locations[minidx]
            distances[minidx] = 99999
            voxel_maps[x, y, z] = chunk_label
        voxel_maps[home_coords[0], home_coords[1], home_coords[2]] = chunk_label
    return voxel_maps


def get_next_home(voxel_maps, XMAX, YMAX, ZMAX):
    """Get the next home coordinates from a current voxel-mapping.

    Parameters
    ----------
    voxel_maps : ndarray
        Volume containing voxel-index mapping to chunks.
    XMAX : int
        Maximum X-coordinate.
    YMAX : int
        Maximum Y-coordinate.
    ZMAX : int
        Maximum Z-coordinate.

    Returns
    -------
    x, y, z : tuple of int
        X/Y/Z coordinates of next home location.

    """
    search_area = voxel_maps.copy()
    search_area = search_area == -1
    x, y, z = 0, 0, 0
    for i in range(ZMAX):
        if np.sum(search_area[:, :, i]) > 0:
            z += i
            break
    for j in range(YMAX):
        if np.sum(search_area[:, j, z]) > 0:
            y += j
            break
    for k in range(XMAX):
        if np.sum(search_area[k, y, z]) > 0:
            x += k
            break
    return x, y, z


def generate_chunk_mask(mask, num_chunks, chunk_size, out_dir, radius=20):
    """Generate chunk index mask.

    Parameters
    ----------
    mask : niimg-like object
        Binary brain volume niimg to divide into chunks
    num_chunks : int
        Number of chunks to divide brain volume into.
    chunk_size : int
        Number of voxels in each chunk.
    out_dir : str
        Directory to save chunk index mask to.
    radius : int, optional
        Radius to search for unassigned voxels, by default 20.

    """
    XMAX, YMAX, ZMAX = mask.shape
    voxel_map = mask.get_fdata().astype(int).copy() * -1
    for i in trange(num_chunks):
        home = get_next_home(voxel_map, XMAX, YMAX, ZMAX)
        voxel_map = label_chunk(
            radius, home, voxel_map, i + 1, chunk_size, XMAX, YMAX, ZMAX
        )
    out_fname = os.path.join(os.path.abspath(out_dir), "chunk_idx.nii.gz")
    image.new_img_like(mask, voxel_map).to_filename(out_fname)



"""
Useful utilities for working with the Precomputed Connectome

"""

import os
import csv
import numpy as np
from tqdm import tqdm
from glob import glob
from nilearn import image
from nilearn._utils import check_niimg
from pfctoolkit import datasets, surface


def load_roi(roi_path):
    """Load ROIs from path or CSV.

    Parameters
    ----------
    roi_path : str
        Path to CSV containing paths to NIfTI images OR Path to directory containing
        NIfTI images OR Path to NIfTI image.

    Returns
    -------
    roi_paths : list of str
        List of paths to NIfTI image ROIs.

    """
    if os.path.isdir(roi_path):
        roi_paths = glob(os.path.join(os.path.abspath(roi_path), "*.nii*"))
        if len(roi_paths) == 0:
            raise FileNotFoundError("No NIfTI images found!")
    else:
        roi_extension = os.path.basename(roi_path).split(".")[1:]
        if "csv" in roi_extension:
            with open(roi_path, newline="") as f:
                reader = csv.reader(f)
                data = list(reader)
            roi_paths = [path for line in data for path in line]
        elif "nii" in roi_extension:
            roi_paths = [roi_path]
        else:
            raise ValueError(
                "Input File is not a NIfTI or a CSV containing paths to NIfTIs"
            )
    print(f"Found {len(roi_paths)} ROIs...")
    return roi_paths


def get_chunks(rois, config):
    """Get list of dicts containing Chunks to load and their associated ROIs

    Parameters
    ----------
    rois : list of str
        List of paths to ROIs to find chunks for.
    config : Config
        Configuration object.

    Returns
    -------
    dict of dicts
        Dict of dicts containing Chunk path and list of associated ROI paths.

    """
    chunk_dict = {}
    chunk_map = image.load_img(config.get("chunk_idx"))
    for roi in tqdm(rois):
        roi_image = image.load_img(roi)
        bin_roi_image = image.math_img("img != 0", img=roi_image)
        roi_chunks = image.math_img(
            "img * mask", img=bin_roi_image, mask=chunk_map
        ).get_fdata()
        chunks = np.unique(roi_chunks).astype(int)
        chunks = chunks[chunks != 0]
        for chunk in chunks:
            if chunk in chunk_dict:
                chunk_dict[chunk].append(roi)
            else:
                chunk_dict[chunk] = [roi]
    return chunk_dict


def get_voxel_conn_map(x, y, z, map_type, config):
    """Get connectivity map for single voxel

    Parameters
    ----------
    x : int
        X-coordinate for voxel to retrieve connectivity map for.
    y : int
        Y-coordinate for voxel to retrieve connectivity map for.
    z : int
        Z-coordinate for voxel to retrieve connectivity map for.
    map_type : str
        Type of connectivity map to retrieve: "t", "avgr", or "fz".
    config : Config
        Configuration object.

    Returns
    -------
    Nifti1Image
        Nifti image object.

    """
    chunk_paths = {
        "avgr": config.get("avgr"),
        "fz": config.get("fz"),
        "t": config.get("t"),
        "combo": config.get("combo"),
    }
    chunk_type = {
        "avgr": "AvgR",
        "fz": "AvgR_Fz",
        "t": "T",
        "combo": "Combo",
    }
    chunk_map = image.load_img(config.get("chunk_idx"))
    brain_mask = datasets.get_img(config.get("mask"))
    affine = brain_mask.affine
    mni_coordinate = np.array([[x, y, z]])
    voxel_coordinate = (
        np.linalg.inv(affine).dot(np.append(mni_coordinate, 1).T).astype(int)[:3]
    )

    voxel_data = np.zeros(brain_mask.shape)
    voxel_data[voxel_coordinate[0], voxel_coordinate[1], voxel_coordinate[2]] = 1
    voxel_img = image.new_img_like(brain_mask, voxel_data)

    roi_chunks = image.math_img("img * mask", img=voxel_img, mask=chunk_map).get_fdata()
    chunks = np.unique(roi_chunks).astype(int)
    chunk = chunks[chunks != 0][0]

    image_type = config.get("type")
    if image_type == "volume":
        brain_masker = NiftiMasker(datasets.get_img(config.get("mask")))
        chunk_masker = NiftiMasker(
            image.math_img(f"img=={chunk}", img=config.get("chunk_idx"))
        )
    elif image_type == "surface":
        brain_masker = surface.GiftiMasker(datasets.get_img(config.get("mask")))
        chunk_masker = surface.GiftiMasker(
            surface.new_gifti_image(
                datasets.get_img(config.get("chunk_idx")).agg_data() == chunk
            )
        )
    chunk_data = np.load(
        os.path.join(chunk_paths[map_type], f"{chunk}_{chunk_type[map_type]}.npy")
    )
    voxel_connectivity = np.dot(chunk_masker.transform(voxel_img), chunk_data)
    voxel_connectivity_img = brain_masker.inverse_transform(voxel_connectivity)

    return voxel_connectivity_img


class NiftiMasker:
    """A Faster NiftiMasker.

    Attributes
    ----------
    mask_img : nibabel.nifti1.Nifti1Image
        Nifti binary mask.
    mask_idx : numpy.ndarray
        1D numpy.ndarray containing indexes from flattened mask image.
    mask_shape : tuple of ints
        Shape of mask_idx.
    mask_size : int
        Number of voxels in entire space, including outside the brain.

    """

    def __init__(self, mask_img=None):
        """

        Parameters
        ----------
        mask_img : Niimg-like object
            If string, consider it as a path to NIfTI image and call `nibabel.load()` on
            it. The '~' symbol is expanded to the user home folder. If it is an object,
            check if affine attribute is present, raise `TypeError` otherwise.

        """
        self.mask_img = check_niimg(mask_img)
        self.mask_data = self.mask_img.get_fdata().astype(np.float32)
        (self.mask_idx,) = np.where((self.mask_data != 0).flatten())
        self.mask_shape = self.mask_data.shape
        self.mask_size = np.prod(self.mask_shape)

    def transform(self, niimg=None, weight=False):
        """Masks 3D Nifti file into 1D array. Retypes to float32.

        Parameters
        ----------
        niimg : Niimg-like object
            If string, consider it as a path to NIfTI image and call `nibabel.load()` on
            it. The '~' symbol is expanded to the user home folder. If it is an object,
            check if affine attribute is present, raise `TypeError` otherwise.
        weight : bool, default False
            If True, transform the niimg with weighting. If False, transform the niimg
            without weighting.

        Returns
        -------
        region_signals : numpy.ndarray
            Masked Nifti file.

        """
        niimg = np.nan_to_num(check_niimg(niimg).get_fdata()).astype(np.float32)
        if weight:
            img = np.multiply(self.mask_data, niimg)
        else:
            img = niimg
        return np.take(img.flatten(), self.mask_idx)

    def inverse_transform(self, flat_niimg=None):
        """Unmasks 1D array into 3D Nifti file. Retypes to float32.

        Parameters
        ----------
        flat_niimg : numpy.ndarray
            1D array to unmask.

        Returns
        -------
        niimg : nibabel.nifti1.Nifti1Image
            Unmasked Nifti.

        """
        new_img = np.zeros(self.mask_size, dtype=np.float32)
        new_img[self.mask_idx] = flat_niimg.astype(np.float32)
        return image.new_img_like(self.mask_img, new_img.reshape(self.mask_shape))

    def mask(self, niimg=None):
        """Masks 3D Nifti file into Masked 3D Nifti file.

        Parameters
        ----------
        niimg : Niimg-like object
            If string, consider it as a path to NIfTI image and call `nibabel.load()`
            on it. The '~' symbol is expanded to the user home folder. If it is an
            object, check if affine attribute is present, raise `TypeError` otherwise.

        Returns
        -------
        masked_niimg : nibabel.nifti1.Nifti1Image
            Masked Nifti file.

        """
        return self.inverse_transform(self.transform(niimg))



"""
Datasets used with the Precomputed Connectome

"""
import nibabel as nib
import importlib.resources as pkg_resources

datasets = {
    "MNI152_T1_2mm_brain": "MNI152_T1_2mm_brain.nii.gz",
    "MNI152_T1_2mm_brain_mask": "MNI152_T1_2mm_brain_mask.nii.gz",
    "MNI152_T1_2mm_brain_mask_dil": "MNI152_T1_2mm_brain_mask_dil.nii.gz",
    "MNI152_T1_1mm_brain": "MNI152_T1_1mm_brain.nii.gz",
    "MNI152_T1_1mm_brain_mask": "MNI152_T1_1mm_brain_mask.nii.gz",
    "MNI152_T1_1mm_brain_mask_dil": "MNI152_T1_1mm_brain_mask_dil.nii.gz",
    "fs5_mask": "fs5_mask.gii",
    "fs5_mask_lh": "fs5_mask_lh.gii",
    "fs5_mask_rh": "fs5_mask_rh.gii",
}


def get_img(ds):
    """Get a standard image file as a Niimg

    Parameters
    ----------
    ds : str
        Name of image to get.\n
        Volume Masks:\n
        "MNI152_T1_2mm_brain"\n
        "MNI152_T1_2mm_brain_mask"\n
        "MNI152_T1_2mm_brain_mask_dil"\n
        "MNI152_T1_1mm_brain"\n
        "MNI152_T1_1mm_brain_mask"\n
        "MNI152_T1_1mm_brain_mask_dil"\n
        Surface Masks:\n
        "fs5_mask"\n
        "fs5_mask_lh"\n
        "fs5_mask_rh"\n

    Returns
    -------
    Niimg-like object

    """
    assert ds in datasets.keys(), "Unknown image specified"
    fname = datasets[ds]
    from . import data

    with pkg_resources.path(data, fname) as datafile:
        return nib.load(str(datafile))


def get_img_path(ds):
    """Get a standard image file path

    Parameters
    ----------
    ds : str
        Name of image to get path of.\n
        Volume Masks\n
        "MNI152_T1_2mm_brain"\n
        "MNI152_T1_2mm_brain_mask"\n
        "MNI152_T1_2mm_brain_mask_dil"\n
        "MNI152_T1_1mm_brain"\n
        "MNI152_T1_1mm_brain_mask"\n
        "MNI152_T1_1mm_brain_mask_dil"\n
        Surface Masks\n
        "fs5_mask"\n
        "fs5_mask_lh"\n
        "fs5_mask_rh"\n

    Returns
    -------
    str
        Path to image.

    """
    assert ds in datasets.keys(), "Unknown image specified"
    fname = datasets[ds]
    from . import data

    with pkg_resources.path(data, fname) as datafile:
        return str(datafile)



"""
Configuration objects to facilitate using the Precomputed Connectome.

"""

import os
import json


class Config:
    def __init__(self, pcc):
        """Generate Config object.

        Parameters
        ----------
        pcc : str
            Name of configfile. Must exist in $HOME/pfctoolkit_config/.

        Raises
        ------
        OSError
            Config is unable to be used because of missing resources.
        FileNotFoundError
            Requsted config file does not exist.
        """
        home = os.path.expanduser('~')
        configfile = os.path.join(home, f"pfctoolkit_config/{pcc}.json")
        try:
            with open(str(configfile)) as js:
                self.config = json.load(js)
            if self.check():
                print(f"Config {self.config['name']} loaded")
            else:
                raise OSError(f"Config {self.config['name']} unavailable")
        except FileNotFoundError:
            raise FileNotFoundError(f"PCC config file {configfile} does not exist!")

    def check(self):
        """Check that all resources specified in config file are accessible and exist.

        Returns
        -------
        bool
            If True, then all resources specified in config file are accessible and
            exist.
        """
        checks = ["avgr", "fz", "t", "combo", "std", "norm", "chunk_idx"]
        return all(list(map(os.path.exists, [self.config[key] for key in checks])))

    def get(self, key):
        """Get value in config specified by key.

        Parameters
        ----------
        key : str
            Key of value to retrieve from config.

        Returns
        -------
        str
            Value of Key in config.
        """
        return self.config[key]

    def get_config(self):
        """Get config as dict.

        Returns
        -------
        dict
            Configuration as dict.
        """
        return self.config



"""Tools for Surface Gifti files

"""

import os
import csv
import numpy as np
import nibabel as nib
from pfctoolkit import datasets


class GiftiMasker:
    def __init__(self, mask_img):
        """

        Parameters
        ----------
        mask_img : Giimg-like object
            If str, path to binary gifti mask. If giftiimg, gifti image object. Assumes
            the gifti only contains a single 1-D data array.

        """
        if type(mask_img) == str:
            self.mask_img = nib.load(mask_img)
        else:
            self.mask_img = mask_img
        self.mask_data = self.mask_img.agg_data()
        (self.mask_idx,) = np.where(self.mask_data != 0)
        self.mask_shape = self.mask_data.shape
        self.mask_size = np.prod(self.mask_shape)

    def transform(self, giimg=None, weight=False):
        """Masks Gifti file into 1D array. Retypes to float32.

        Parameters
        ----------
        giimg : Giimg-like object
            If string, consider it as a path to GIfTI image and call `nibabel.load()` on
            it. The '~' symbol is expanded to the user home folder. If it is an object,
            check if affine attribute is present, raise `TypeError` otherwise. If
            ndarray, consider it as a gifti data array.
        weight : bool, default False
            If True, transform the niimg with weighting. If False, transform the niimg
            without weighting.

        Returns
        -------
        region_signals : numpy.ndarray
            Masked Nifti file.

        """
        if type(giimg) == str:
            giimg_data = nib.load(giimg).agg_data().astype(np.float32)
        elif type(giimg) == np.ndarray:
            giimg_data = giimg
        else:
            giimg_data = giimg.agg_data().astype(np.float32)
        if weight:
            img = np.multiply(self.mask_data, giimg_data)
        else:
            img = giimg_data
        return np.take(img, self.mask_idx)

    def inverse_transform(self, flat_giimg=None):
        """Unmasks 1D array into 3D Gifti file. Retypes to float32.

        Parameters
        ----------
        flat_giimg : numpy.ndarray
            1D array to unmask.

        Returns
        -------
        giimg : nibabel.gifti.gifti.GiftiImage
            Unmasked Gifti.

        """
        new_img = np.zeros(self.mask_size, dtype=np.float32)
        new_img[self.mask_idx] = flat_giimg.astype(np.float32)
        return new_gifti_image(data=new_img)

    def mask(self, giimg=None):
        """Masks 3D Gifti file into Masked Gifti file.

        Parameters
        ----------
        giimg : Giimg-like object
            If string, consider it as a path to GIfTI image and call `nibabel.load()` on
            it. The '~' symbol is expanded to the user home folder. If it is an object,
            check if affine attribute is present, raise `TypeError` otherwise.

        Returns
        -------
        masked_giimg : nibabel.gifti.gifti.GiftiImage
            Masked Gifti image.

        """
        return self.inverse_transform(self.transform(giimg))


def new_gifti_image(data, intent=0, datatype=16, metadata=None):
    """NiBabel wrapper to generate a gifti image with data array and metadata.

    Parameters
    ----------
    data : ndarray
        1-D ndarray containing one hemisphere surface data.
    intent : int
        Intent code for Gifti File. Defaults to 0 (Intent = NONE).\n
        Available intent codes:\n
        NIFTI_INTENT_NONE - 0\n
        NIFTI_INTENT_CORREL - 2\n
        NIFTI_INTENT_TTEST - 3\n
        NIFTI_INTENT_ZSCORE - 5\n
        NIFTI_INTENT_PVAL - 22\n
        NIFTI_INTENT_LOGPVAL - 23\n
        NIFTI_INTENT_LOG10PVAL - 24\n
        NIFTI_INTENT_LABEL - 1002\n
        NIFTI_INTENT_POINTSET - 1008\n
        NIFTI_INTENT_TRIANGLE - 1009\n
        NIFTI_INTENT_TIME_SERIES - 2001\n
        NIFTI_INTENT_NODE_INDEX - 2002\n
        NIFTI_INTENT_SHAPE - 2005\n
        More intent codes can be found at: https://nifti.nimh.nih.gov/nifti-1/documentation/nifti1fields/nifti1fields_pages/group__NIFTI1__INTENT__CODES.html
    datatype : int
        Datatype for gifti image. Defaults to 16 (dtype = float32)\n
        Available datatypes:\n
        UINT8 - 2\n
        INT32 - 8\n
        FLOAT32 - 16\n
    metadata : dict
        Dictionary of metadata for gifti image.

    Returns
    -------
    nibabel.gifti.gifti.GiftiImage
        Gifti image with specified metadata and data array.

    """
    dtypes = {2: np.uint8, 8: np.int32, 16: np.float32}
    data = data.astype(dtypes[datatype])
    if metadata:
        metadata = nib.gifti.GiftiMetaData.from_dict(metadata)
    gifti_data = nib.gifti.GiftiDataArray(data=data, intent=intent, datatype=datatype)
    gifti_img = nib.gifti.GiftiImage(meta=metadata, darrays=[gifti_data])
    return gifti_img


def concat_hemispheres_to_csv(gifti_paths, output_dir="", mask=""):
    """Concatenate a list of giftis together into a csv to construct a data matrix for
    use with PALM. Assumes that input giftis contain a single 1-D data array with
    functional/statistical data for surface mesh vertices.

    Parameters
    ----------
    gifti_paths : str
        Path to a two-column CSV of paths to giftis where each row corresponds to a
        subject, the first column corresponds to the subject's Left Hemisphere ROI mask,
        and the second column corresponds to the subject's Right Hemisphere ROI mask.\n
        CSV Format:\n
        /path/to/subject1/lh_roi.nii.gz,/path/to/subject1/rh_roi.nii.gz\n
        /path/to/subject2/lh_roi.nii.gz,/path/to/subject2/rh_roi.nii.gz\n
        /path/to/subject3/lh_roi.nii.gz,/path/to/subject3/rh_roi.nii.gz\n
    output_dir : str, optional
        Output directory, by default "". If not specified, generated data matrix csv
        will be output to same directory as input csv.
    mask : str, {'fs5_mask'}, optional
        Mask name (from nimlab.datasets). Defaults to no masking. The provided
        fsaverage5 mask is created from Ryan Darby's ADNI data.

    """
    roi_files = []
    flist = open(gifti_paths)
    reader = csv.reader(flist, delimiter=",")
    for f in reader:
        roi_files.append(f)
    flist.close()
    if mask:
        masker = GiftiMasker(datasets.get_img(mask))
    if output_dir == "":
        output_dir = os.path.dirname(gifti_paths)
    subject_data = []
    for subject in roi_files:
        lh = nib.load(subject[0]).agg_data()
        rh = nib.load(subject[1]).agg_data()
        data = np.concatenate((lh, rh))
        if mask:
            data = masker.transform(data)
        subject_data.append(data)
    fname = os.path.join(output_dir, "data.csv")
    np.savetxt(fname, np.stack(subject_data), delimiter=",")



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

LICENSE = "BSD License"

setuptools.setup(
    name="pfc-toolkit",
    version="2024.2.15.1",
    author="William Drew",
    author_email="william.drew100@gmail.com",
    description="The Precomputed Functional Connectome Toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/thewilliamdrew/pfc-toolkit",
    project_urls={
        "Bug Tracker": "https://github.com/thewilliamdrew/pfc-toolkit/issues",
    },
    license=LICENSE,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    include_package_data=True,
    package_data={
        "pfctoolkit.chunks": ["*.nii.gz", "*.gii"],
        "pfctoolkit.data": ["*.nii.gz", "*.gii"],
        "pfctoolkit.configs": ["*.json"],
        "pfctoolkit.scripts": ["*.py"],
    },
    scripts=[
        "src/pfctoolkit/scripts/connectome_precomputed",
        "src/pfctoolkit/scripts/generate_pfc_combo_chunks",
        "src/pfctoolkit/scripts/generate_pfc_fc_chunks",
        "src/pfctoolkit/scripts/generate_pfc_weighted_masks",
    ],
    install_requires=[
        "tqdm",
        "numba",
        "numpy",
        "scipy",
        "nibabel",
        "nilearn",
        "natsort",
        "importlib",
    ],
    python_requires=">=3.6",
)



"""
Tools to perform Lesion Network Mapping with the Precomputed Connectome

"""
import os
import numpy as np
import nibabel as nib
from numba import jit
from nilearn import image
from pfctoolkit import tools, datasets, surface


def process_chunk(chunk, rois, config):
    """Compute chunk contribution to FC maps for a given list of ROIs.

    Parameters
    ----------
    chunk : int
        Index of chunk to be processed.
    rois : list of str
        List of ROI paths to be processed.
    config : pfctoolkit.config.Config
        Configuration of the precomputed connectome.

    Returns
    -------
    contributions : dict of ndarray
        Dictionary containing contributions to network maps.

    """
    chunk_paths = {
        "avgr": config.get("avgr"),
        "fz": config.get("fz"),
        "t": config.get("t"),
        "combo": config.get("combo"),
    }
    image_type = config.get("type")
    if image_type == "volume":
        brain_masker = tools.NiftiMasker(datasets.get_img(config.get("mask")))
        chunk_masker = tools.NiftiMasker(
            image.math_img(f"img=={chunk}", img=config.get("chunk_idx"))
        )
    elif image_type == "surface":
        brain_masker = surface.GiftiMasker(datasets.get_img(config.get("mask")))
        chunk_masker = surface.GiftiMasker(
            surface.new_gifti_image(
                datasets.get_img(config.get("chunk_idx")).agg_data() == chunk
            )
        )
    brain_weights = np.array([brain_masker.transform(roi) for roi in rois])
    chunk_weights = np.array([chunk_masker.transform(roi) for roi in rois])
    brain_masks = np.array([(weights != 0) for weights in brain_weights])
    chunk_masks = np.array([(weights != 0) for weights in chunk_weights])
    norm_weight = chunk_masker.transform(config.get("norm"))
    std_weight = chunk_masker.transform(config.get("std"))

    norm_chunk_masks, std_chunk_masks = compute_chunk_masks(
        chunk_weights, norm_weight, std_weight
    )
    contributions = {}
    for chunk_type in [
        ("avgr", "AvgR"),
        ("fz", "AvgR_Fz"),
        ("t", "T"),
        ("combo", "Combo"),
    ]:
        chunk_data = np.load(
            os.path.join(chunk_paths[chunk_type[0]], f"{chunk}_{chunk_type[1]}.npy")
        )
        expected_shape = (config.get("chunk_size"), config.get("brain_size"))
        if chunk_data.shape != expected_shape:
            raise TypeError(
                f"Chunk expected to have shape {expected_shape} but"
                f" instead has shape {chunk_data.shape}!"
            )
        if chunk_type[0] == "combo":
            numerator = compute_numerator(norm_chunk_masks)
            for i, roi in enumerate(rois):
                denominator = compute_denominator(
                    brain_weights,
                    chunk_weights,
                    brain_masks,
                    chunk_masks,
                    chunk_data,
                    i,
                )
                contributions[roi]["numerator"] = numerator[i]
                contributions[roi]["denominator"] = denominator
        else:
            network_maps = compute_network_maps(std_chunk_masks, chunk_data)
            for i, roi in enumerate(rois):
                if chunk_type[0] == "avgr":
                    contributions[roi] = {
                        chunk_type[0]: network_maps[i, :],
                    }
                else:
                    contributions[roi][chunk_type[0]] = network_maps[i, :]
    network_weights = compute_network_weights(std_chunk_masks)
    for i, roi in enumerate(rois):
        contributions[roi]["network_weight"] = network_weights[i]
    return contributions


def update_atlas(contribution, atlas):
    """Update atlas with chunk contributions.

    Parameters
    ----------
    contribution : dict
        dict containing FC and scaling factor contributions from a chunk.
    atlas : dict
        dict containing in-progress FC maps and scaling factor accumulators.

    Returns
    -------
    atlas : dict
        Updated dict containing in-progress FC maps and scaling factor
        accumulators following consolidation of the contribution.

    """
    for roi in contribution.keys():
        if roi in atlas:
            atlas[roi]["avgr"] += contribution[roi]["avgr"]
            atlas[roi]["fz"] += contribution[roi]["fz"]
            atlas[roi]["t"] += contribution[roi]["t"]
            atlas[roi]["network_weight"] += contribution[roi]["network_weight"]
            atlas[roi]["numerator"] += contribution[roi]["numerator"]
            atlas[roi]["denominator"] += contribution[roi]["denominator"]
        else:
            atlas[roi] = {
                "avgr": contribution[roi]["avgr"],
                "fz": contribution[roi]["fz"],
                "t": contribution[roi]["t"],
                "network_weight": contribution[roi]["network_weight"],
                "numerator": contribution[roi]["numerator"],
                "denominator": contribution[roi]["denominator"],
            }
    return atlas


def publish_atlas(atlas, output_dir, config):
    """Runs final computation on the atlas and outputs network maps to file.

    Parameters
    ----------
    atlas : dict
        dict containing in-progress FC maps and scaling factor accumulators.
    output_dir : str
        Output directory.
    config : pfctoolkit.config.Config
        Configuration of the precomputed connectome.

    """
    output_dir = os.path.abspath(output_dir)
    image_type = config.get("type")
    if image_type == "volume":
        brain_masker = tools.NiftiMasker(datasets.get_img(config.get("mask")))
        extension = ".nii.gz"
    elif image_type == "surface":
        brain_masker = surface.GiftiMasker(datasets.get_img(config.get("mask")))
        extension = ".gii"
    for roi in atlas:
        atlas[roi]["denominator"] = final_denominator(atlas[roi]["denominator"])
        atlas[roi]["avgr"] = atlas[roi]["avgr"] / atlas[roi]["network_weight"]
        atlas[roi]["fz"] = atlas[roi]["fz"] / atlas[roi]["network_weight"]
        atlas[roi]["t"] = atlas[roi]["t"] / atlas[roi]["network_weight"]
        scaling_factor = atlas[roi]["numerator"] / atlas[roi]["denominator"]
        subject_name = os.path.basename(roi).split(".nii")[0]
        for map_type in [("avgr", "AvgR"), ("fz", "AvgR_Fz"), ("t", "T")]:
            output_fname = f"{subject_name}_Precom_{map_type[1]}{extension}"
            output_path = os.path.join(output_dir, output_fname)
            atlas[roi][map_type[0]] = atlas[roi][map_type[0]] * scaling_factor
            out_img = brain_masker.inverse_transform(atlas[roi][map_type[0]])
            out_img.to_filename(output_path)
    print(f"Network maps output to {output_dir}")
    print("Done!")


@jit(nopython=True)
def final_denominator(denominator):
    return np.sqrt(denominator)


@jit(nopython=True)
def compute_network_weights(std_chunk_masks):
    """Compute network weights.

    Parameters
    ----------
    std_chunk_masks : ndarray
        Chunk-masked ROIs weighted by BOLD standard deviation.

    Returns
    -------
    network_weights : ndarray
        Contribution to total network map weights.

    """
    network_weights = np.sum(std_chunk_masks, axis=1)
    return network_weights


@jit(nopython=True)
def compute_network_maps(std_chunk_masks, chunk_data):
    """Compute network maps.

    Parameters
    ----------
    std_chunk_masks : ndarray
        Chunk-masked ROIs weighted by BOLD standard deviation.
    chunk_data : ndarray
        Chunk data.

    Returns
    -------
    network maps : ndarray
        Network map contributions from chunk.

    """
    network_maps = np.dot(std_chunk_masks, chunk_data)
    return network_maps


@jit(nopython=True)
def compute_denominator(
    brain_weights, chunk_weights, brain_masks, chunk_masks, chunk_data, i
):
    """Compute denominator contribution.

    Parameters
    ----------
    brain_weights : ndarray
        Brain-masked weighted ROIs.
    chunk_weights : ndarray
        Chunk-masked weighted ROIs.
    brain_masks : ndarray
        Brain-masked unweighted ROIs.
    chunk_masks : ndarray
        Chunk-masked unweighted ROIs.
    chunk_data : ndarray
        Chunk data.
    i : int
        Index of processed ROI.

    Returns
    -------
    denominator : float
        Contribution to denominator.

    """
    chunk_masked = np.multiply(
        np.reshape(chunk_weights[i][chunk_masks[i]], (-1, 1)),
        chunk_data[chunk_masks[i], :],
    )
    brain_masked = np.multiply(
        brain_weights[i][brain_masks[i]], chunk_masked[:, brain_masks[i]]
    )
    denominator = np.sum(brain_masked)
    return denominator


@jit(nopython=True)
def compute_numerator(norm_chunk_masks):
    """Compute numerator contribution.

    Parameters
    ----------
    norm_chunk_masks : ndarray
        ROI chunk masks weighted with BOLD norms.

    Returns
    -------
    numerator : float
        Numerator contribution

    """
    numerator = np.sum(norm_chunk_masks, axis=1)
    return numerator


@jit(nopython=True)
def compute_chunk_masks(chunk_weights, norm_weight, std_weight):
    """Compute weighted chunk masks.

    Parameters
    ----------
    chunk_weights : ndarray
        Chunk-masked weighted ROIs.
    norm_weight : ndarray
        Chunk-masked voxel BOLD norms.
    std_weight : ndarray
        Chunk-masked voxel BOLD standard deviations.

    Returns
    -------
    norm_weighted_chunk_masks : ndarray
        ROI chunk masks weighted with BOLD norms.
    std_weighted_chunk_masks : ndarray
        ROI chunk masks weighted with BOLD standard deviations.

    """
    norm_weighted_chunk_masks = np.multiply(chunk_weights, norm_weight)
    std_weighted_chunk_masks = np.multiply(chunk_weights, std_weight)
    return norm_weighted_chunk_masks, std_weighted_chunk_masks



"""
Tools to generate a Precomputed Functional Connectome.

"""

import os
import time
import numpy as np
import nibabel as nib
from glob import glob
from tqdm import tqdm
from numba import jit
from pathlib import Path
from nilearn import image
from pfctoolkit import tools
from pfctoolkit import surface
from natsort import natsorted


@jit(nopython=True)
def extract_chunk_signals(connectome_mat, roi_mat):
    """Extract Chunk TC Signal from a connectome subject.

    Parameters
    ----------
    connectome_mat : ndarray
        Connectome subject TC matrix.
    roi_mat : ndarray
        Chunk ROI binary mask

    Returns
    -------
    roi_masked_tc : ndarray
        TC signals extracted from connectome TC according to ROI binary mask

    """
    roi_masked_tc = connectome_mat[:, roi_mat > 0]
    return roi_masked_tc


@jit(nopython=True)
def dot(a, b):
    """Compute the dot product of `a` and b`"""
    return np.dot(a, b)


@jit(nopython=True)
def divide(a, b):
    """Compute the quotient of `a` and `b`."""
    return np.divide(a, b)


@jit(nopython=True)
def arctanh(a):
    """Compute the hyperbolic arctangent."""
    return np.arctanh(a)


@jit(nopython=True)
def tanh(a):
    """Compute the tangent."""
    return np.tanh(a)


@jit(nopython=True)
def subtract(a, b):
    """Compute the difference between `a` and `b`."""
    return np.subtract(a, b)


@jit(nopython=True)
def multiply(a, b):
    """Compute the product of `a` and `b`."""
    return np.multiply(a, b)


@jit(nopython=True)
def sqrt(a):
    """Compute the square root."""
    return np.sqrt(a)


@jit(nopython=True)
def make_combo_chunk(agg_combo_chunk, chunk_bold, bold):
    """Compute a step of a combo chunk.

    Parameters
    ----------
    agg_combo_chunk : ndarray
        Aggregate Combo chunk.
    chunk_bold : ndarray
        Chunk masked BOLD signal.
    bold : ndarray
        Whole brain BOLD signal.

    Returns
    -------
    ndarray
        Next step of the aggregate combo chunk calculation.
    """
    return np.add(agg_combo_chunk, np.dot(chunk_bold.T, bold))


def make_fz_maps(connectome_files, roi_mat):
    """Make Fz Maps for a chunk ROI and a connectome subject.

    Parameters
    ----------
    connectome_files : (str, str)
        Tuple of paths for connectome subject TC matrix and TC norm vector.
    roi_mat : ndarray
        Brain-masked binary chunk ROI with shape (<size of brain>,)

    """
    connectome_mat = np.load(connectome_files[0]).astype(np.float32)
    connectome_norms_mat = np.load(connectome_files[1]).astype(np.float32)
    chunk_tc = extract_chunk_signals(connectome_mat, roi_mat)
    corr_num = dot(connectome_mat.T, chunk_tc)
    corr_denom = dot(
        connectome_norms_mat.reshape(-1, 1),
        np.linalg.norm(chunk_tc, axis=0).reshape(1, -1),
    )
    np.seterr(invalid="ignore")
    corr = divide(corr_num, corr_denom)
    corr[np.isnan(corr)] = 0
    fz = arctanh(corr)
    # Fix infinite values and nans in the case of single voxel autocorrelations
    finite_max = np.amax(np.ma.masked_invalid(fz), 1).data
    rows, cols = np.where(np.isinf(fz) | np.isnan(fz))
    for row, col in tqdm(zip(rows, cols)):
        fz[row, col] = finite_max[row]
    return fz


@jit(nopython=True)
def welford_update_map(count, mean, M2, newMap):
    """Update a Welford map with data from a new map. See
    https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm

    Parameters
    ----------
    count : int
        Welford counter.
    mean : ndarray, shape (<brain size>, <ROI size>)
        Mean aggregate.
    M2 : ndarray, shape (<brain size>, <ROI size>)
        M2 aggregate.
    newMap : ndarray, shape (<brain size>, <ROI size>)
        New data to incorporate into a Welford map.

    Returns
    -------
    updatedAggregateMap : ndarray
        Updated Welford map.

    """
    count += 1
    delta = np.subtract(newMap, mean)
    mean += np.divide(delta, count)
    delta2 = np.subtract(newMap, mean)
    M2 += np.multiply(delta, delta2)
    return count, mean, M2


@jit(nopython=True)
def welford_finalize_map(count, mean, M2):
    """Convert a Welford map into maps of arrays containing the statistics
    [mean, variance, sampleVariancce].

    Parameters
    ----------
    count : int
        Welford counter.
    mean : ndarray
        Welford Mean array.
    M2 : ndarray
        Welford M2 array.

    Returns
    -------
    ndarray, ndarray
        Array of Welford means and sample variances.

    """
    sampleVariance = np.divide(M2, count - 1)
    return mean, sampleVariance


def make_stat_maps(count, mean, M2, output_dir, chunk_idx):
    """Generate statistical maps from welford maps and output to file.

    Parameters
    ----------
    count : int
        Number of connectome subjects.
    mean : ndarray
        Welford Mean array.
    M2 : ndarray
        Welford M2 array.
    output_dir : str
        Path to output directory.
    chunk_idx : int
        Chunk index number being processed.

    """
    mean, sampleVariance = welford_finalize_map(count, mean, M2)
    map_types = ["AvgR_Fz", "AvgR", "T"]
    output_dir = os.path.abspath(output_dir)
    output_dirs = {}
    for map_type in map_types:
        Path(os.path.join(output_dir, map_type)).mkdir(parents=True, exist_ok=True)
        output_dirs[map_type] = os.path.join(
            output_dir, map_type, f"{chunk_idx}_{map_type}.npy"
        )
    print("Output Chunk files to:")
    # Save AvgR_Fz chunk
    np.save(output_dirs["AvgR_Fz"], mean.T)
    print(f"AvgR_Fz: {output_dirs['AvgR_Fz']}")
    # Save AvgR chunk
    np.save(output_dirs["AvgR"], tanh(mean).T)
    print(f"AvgR: {output_dirs['AvgR']}")
    # Save T Chunk
    ttest_denom = sqrt(divide(sampleVariance, count))
    np.save(output_dirs["T"], divide(mean, ttest_denom).T)
    print(f"T: {output_dirs['T']}")


@jit(nopython=True)
def calculate_norm_square(agg_norm_square, bold):
    """Calculate norm square.

    Parameters
    ----------
    agg_norm_square : ndarray
        Aggregate norm square vector.
    bold : ndarray
        BOLD array from a subject to be incorporated into the aggregate norm
        square vector.

    Returns
    -------
    agg_norm_square : ndarray
        Aggregate norm square vector with contribution from BOLD subject.

    """
    return np.add(agg_norm_square, np.sum(np.square(bold), axis=0))


def precomputed_connectome_fc_chunk(
    mask, chunk_idx_mask, chunk_idx, connectome_dir, output_dir
):
    """Generate a precomputed connectome FC map chunk (AvgR/Fz/T).

    Parameters
    ----------
    mask : str
        Path to binary mask (.nii or .gii).
    chunk_idx_mask : str
        Path to mask containing voxel-wise chunk labels (.nii or .gii).
    chunk_idx : int
        Index of chunk to process.
    connectome_dir : str
        Path to individual subject connectome files (dir containing .npy).
    output_dir : str
        Path to output directory.

    """
    start = time.time()
    # Check that mask and chunk_idx_mask are in same space
    if ".nii" in chunk_idx_mask:
        image_type = "volume"
    elif ".gii" in chunk_idx_mask:
        image_type = "surface"
    else:
        raise TypeError("Chunk Index file is not .nii or .gii.")
    if image_type == "volume":
        same_size = (
            (image.get_data(mask) == 1) == (image.get_data(chunk_idx_mask) > 0)
        ).all
    elif image_type == "surface":
        same_size = (
            (nib.load(mask).agg_data() == 1)
            == (nib.load(chunk_idx_mask).agg_data() > 0)
        ).all
    assert same_size, "Binary mask and chunk idx mask do not match!"

    if image_type == "volume":
        masker = tools.NiftiMasker(mask)
        brain_size = int(np.sum(image.get_data(mask)))
    elif image_type == "surface":
        masker = surface.GiftiMasker(mask)
        brain_size = int(np.sum(nib.load(mask).agg_data()))

    # Get list of connectome files
    connectome_files_norms = natsorted(
        glob(os.path.join(connectome_dir, "*_norms.npy"))
    )
    connectome_files = [
        (glob(f.split("_norms")[0] + ".npy")[0], f) for f in connectome_files_norms
    ]
    if len(connectome_files) == 0:
        raise ValueError("No connectome files found")
    else:
        print(f"Found {len(connectome_files)} connectome subjects!")

    # Get binary chunk roi in brain-space
    chunk_roi = masker.transform(chunk_idx_mask) == chunk_idx
    chunk_size = int(np.sum(chunk_roi))

    # Initialize Welford Maps
    count = 0
    mean = np.zeros((brain_size, chunk_size), dtype=np.float32)
    M2 = np.zeros((brain_size, chunk_size), dtype=np.float32)

    # For each connectome subject
    for connectome_file in tqdm(connectome_files):
        # Calculate Fz maps from a connectome subject
        fz_welford = make_fz_maps(connectome_file, chunk_roi)
        # Update Welford Maps
        count, mean, M2 = welford_update_map(count, mean, M2, fz_welford)

    # Finalize Welford Maps and output to file
    print("Outputting Chunk to disk...")
    make_stat_maps(count, mean, M2, output_dir, chunk_idx)
    end = time.time()
    print(f"Elapsed time: {end-start} seconds")


def precomputed_connectome_combo_chunk(
    mask, chunk_idx_mask, chunk_idx, connectome_dir, output_dir
):
    """Generate a precomputed connectome combo chunk.

    Parameters
    ----------
    mask : str
        Path to binary mask (.nii or .gii).
    chunk_idx_mask : str
        Path to mask containing voxel-wise chunk labels (.nii or .gii).
    chunk_idx : int
        Index of chunk to process.
    connectome_dir : str
        Path to individual subject connectome files (dir containing .npy).
    output_dir : str
        Path to output directory.

    """
    start = time.time()
    # Check that mask and chunk_idx_mask are in same space
    if ".nii" in chunk_idx_mask:
        image_type = "volume"
    elif ".gii" in chunk_idx_mask:
        image_type = "surface"
    else:
        raise TypeError("Chunk Index file is not .nii or .gii.")
    if image_type == "volume":
        same_size = (
            (image.get_data(mask) == 1) == (image.get_data(chunk_idx_mask) > 0)
        ).all
    elif image_type == "surface":
        same_size = (
            (nib.load(mask).agg_data() == 1)
            == (nib.load(chunk_idx_mask).agg_data() > 0)
        ).all
    assert same_size, "Binary mask and chunk idx mask do not match!"

    if image_type == "volume":
        masker = tools.NiftiMasker(mask)
        brain_size = int(np.sum(image.get_data(mask)))
    elif image_type == "surface":
        masker = surface.GiftiMasker(mask)
        brain_size = int(np.sum(nib.load(mask).agg_data()))

    # Get list of connectome files
    connectome_files = natsorted(glob(os.path.join(connectome_dir, "*[!_norms].npy")))
    if len(connectome_files) == 0:
        raise ValueError("No connectome files found")
    else:
        print(f"Found {len(connectome_files)} connectome subjects!")

    # Get binary chunk roi in brain-space
    chunk_roi = masker.transform(chunk_idx_mask) == chunk_idx
    chunk_size = int(np.sum(chunk_roi))

    # For each connectome subject aggregate to combo chunk
    agg_combo_chunk = np.zeros((chunk_size, brain_size), dtype=np.float32)
    for subject in tqdm(connectome_files):
        bold = np.load(subject).astype(np.float32)
        assert bold.shape[1] == brain_size, (
            "Mask does not match connectome. "
            f"Mask has size {brain_size} voxels,"
            f" the connectome has {bold.shape[1]}"
            " voxels."
        )
        chunk_bold = bold[:, chunk_roi.astype(bool)]
        agg_combo_chunk = make_combo_chunk(agg_combo_chunk, chunk_bold, bold)

    # Save to file
    Path(os.path.join(output_dir, "Combo")).mkdir(parents=True, exist_ok=True)
    output_dir = os.path.join(
        os.path.abspath(output_dir), "Combo", f"{chunk_idx}_Combo.npy"
    )
    np.save(output_dir, agg_combo_chunk)
    end = time.time()
    print(f"Elapsed time: {end-start} seconds")


def precomputed_connectome_weighted_masks(
    mask, connectome_dir, output_dir, connectome_name=""
):
    """Generate the precomputed connectome norm and stdev weighted masks.

    Parameters
    ----------
    mask : str
        Path to binary mask (.nii or .gii).
    connectome_dir : str
        Path to individual subject connectome files (dir containing .npy).
    output_dir : str
        Path to output directory.
    connectome_name : str, default None
        Name of connectome to use for mask naming. If None, use connectome_dir

    """
    start = time.time()

    # Get image type and load masker
    if ".nii" in mask:
        image_type = "volume"
    elif ".gii" in mask:
        image_type = "surface"
    else:
        raise TypeError("Chunk Index file is not .nii or .gii.")
    if image_type == "volume":
        masker = tools.NiftiMasker(mask)
        brain_size = int(np.sum(image.get_data(mask)))
    elif image_type == "surface":
        masker = surface.GiftiMasker(mask)
        brain_size = int(np.sum(nib.load(mask).agg_data()))

    connectome_dir = os.path.abspath(connectome_dir)
    output_dir = os.path.abspath(output_dir)

    # Load connectome subjects
    connectome_files = natsorted(glob(os.path.join(connectome_dir, "*[!_norms].npy")))
    if len(connectome_files) == 0:
        raise ValueError("No connectome files found")
    else:
        print(f"Found {len(connectome_files)} connectome subjects!")

    # For each connectome subject, compute contribution to norm mask
    agg_norm_square = np.zeros(brain_size, dtype=np.float32)
    timesteps = 0
    for subject in tqdm(connectome_files):
        bold = np.load(subject).astype(np.float32)
        assert bold.shape[1] == brain_size, (
            "Mask does not match connectome. "
            f"Mask has size {brain_size} voxels,"
            f" the connectome has {bold.shape[1]}"
            " voxels."
        )
        agg_norm_square = calculate_norm_square(agg_norm_square, bold)
        timesteps += bold.shape[0]
    if connectome_name == "":
        connectome_name = os.path.basename(connectome_dir)
    if image_type == "volume":
        extension = ".nii.gz"
    elif image_type == "surface":
        extension = ".gii"
    norm_fname = os.path.join(
        output_dir, f"{connectome_name}_norm_weighted_mask" + extension
    )
    std_fname = os.path.join(
        output_dir, f"{connectome_name}_std_weighted_mask" + extension
    )
    norm_mask = sqrt(agg_norm_square)
    masker.inverse_transform(norm_mask).to_filename(norm_fname)
    print(f"Output norm weighted mask to: {norm_fname}")
    std_mask = sqrt(divide(agg_norm_square, timesteps))
    masker.inverse_transform(std_mask).to_filename(std_fname)
    print(f"Output std weighted mask to: {std_fname}")
    end = time.time()
    print(f"Elapsed time: {end-start} seconds")
